/* ***** BEGIN LICENSE BLOCK *****
 *
 * Copyright (c) 2008, Mozilla Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Mozilla Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributor(s):
 *   Josh Aas <josh@mozilla.com>
 *
 * ***** END LICENSE BLOCK ***** */
#include <QWidget>
#include <QPainter>
#include <QColor>
#include <QFont>
#include <QDebug>
#include <QDialog>
#include <QString>
#include <QStyleOptionGraphicsItem>

#include "nptest_platform.h"
#include "npapi.h"
#ifdef MOZ_X11
#include <X11/extensions/shape.h>
#endif
#include "glib.h"
#include <cairo.h>
#include <sys/time.h>
static struct timeval sDiffTime, sCurrTime, sPrevTime;
static int fps_counter = 0;

struct _PlatformData
{
#ifdef MOZ_X11
    Display* display;
    Visual* visual;
    Colormap colormap;
#endif
};
using namespace std;

bool
pluginSupportsWindowMode()
{
    return false;
}

bool
pluginSupportsWindowlessMode()
{
    return true;
}

bool
timeout_data(InstanceData* instanceData)
{
    NPP npp = instanceData->npp;
    if (!npp)
        return false;

    NPWindow& window = instanceData->window;
    NPRect rect;
    rect.left = 0;
    rect.top = 0;
    rect.right = window.width;
    rect.bottom = window.height;
    NPN_InvalidateRect(npp, &rect);
    return true;
}

NPError
pluginInstanceInit(InstanceData* instanceData)
{
    if (instanceData->doPerfTest && !instanceData->sourceTimeout) {
//        instanceData->sourceTimeout = g_timeout_add(35, (GSourceFunc)timeout_data, instanceData);
    }
#ifdef MOZ_X11
    instanceData->platformData = static_cast<PlatformData*>
                                 (NPN_MemAlloc(sizeof(PlatformData)));
    if (!instanceData->platformData)
    {
        printf("NPERR_OUT_OF_MEMORY_ERROR\n");
        return NPERR_OUT_OF_MEMORY_ERROR;
    }

    instanceData->platformData->display = NULL;
    instanceData->platformData->visual = NULL;
    instanceData->platformData->colormap = None;

    return NPERR_NO_ERROR;
#else
    printf("NPERR_INCOMPATIBLE_VERSION_ERROR\n");
    return NPERR_INCOMPATIBLE_VERSION_ERROR;
#endif
    return NPERR_NO_ERROR;
}

void
pluginInstanceShutdown(InstanceData* instanceData)
{
    //printf("QT:	pluginInstanceShutdown\n");
    NPN_MemFree(instanceData->platformData);
    instanceData->platformData = 0;
}

void
pluginDoSetWindow(InstanceData* instanceData, NPWindow* newWindow)
{
    instanceData->window = *newWindow;
}

void
pluginWidgetInit(InstanceData* instanceData, void* oldWindow)
{
    // XXX nothing here yet since we don't support windowed plugins
}

int16_t
pluginHandleEvent(InstanceData* instanceData, void* event)
{

#ifdef MOZ_X11
    XEvent* nsEvent = (XEvent*)event;
    //printf("\nEvent Type %d ", nsEvent->type);
    switch (nsEvent->type)
    {

    case Expose:
    {
        printf("Expose\n");
        break;
    }
    case NoExpose:
    {
        printf("NoExpose\n");
        break;
    }
    case MapRequest:
    {
        printf("MapRequest\n");
        break;
    }
    case ResizeRequest:
    {
        printf("ResizeRequest\n");
        break;
    }
    case ConfigureRequest:
    {
        printf("ConfigureRequest\n");
        break;
    }
    case CirculateRequest:
    {
        printf("Circulate\n");
        break;
    }
    case SelectionClear:
    {
        printf("SelectionClear\n");
        break;
    }
    case GraphicsExpose:
    {
        // printf("GraphicsExpose\n");
        if (!fps_counter)
            gettimeofday(&sPrevTime, 0);
        fps_counter++;
        if (fps_counter > 50) {
            gettimeofday(&sCurrTime, 0);
            timersub(&sCurrTime, &sPrevTime, &sDiffTime);
            float time = fps_counter / (sDiffTime.tv_sec + (float)sDiffTime.tv_usec/1000000);
            if (instanceData->showFps) {
                printf("Expose freq diff:%ld.%06ld fps:%g\n", sDiffTime.tv_sec, sDiffTime.tv_usec, time);
                instanceData->lastFpsNumber = time;
            }
            fps_counter = 0;
        }

        if (((XGraphicsExposeEvent*)event)->display)
            pluginDrawWindow(instanceData, event);
        else
            pluginDrawLocalWindow(instanceData, event);
        timeout_data(instanceData);
        break;
    }
    case MotionNotify:
    {
        printf("MotionNotify\n");
        XMotionEvent* motion = &nsEvent->xmotion;
        instanceData->lastMouseX = motion->x;
        instanceData->lastMouseY = motion->y;
        break;
    }
    case ButtonPress:
    {
        printf("ButtonPress\n");

        QWidget* mQAppWindow = new QWidget;
        mQAppWindow->setWindowTitle("Window spawned by test plugin");
        mQAppWindow->show();

        break;
    }
    case ButtonRelease:
    {
        printf("ButtonRelease\n");
        XButtonEvent* button = &nsEvent->xbutton;
        instanceData->lastMouseX = button->x;
        instanceData->lastMouseY = button->y;
        break;
    }
    default:
        break;
    }
#endif
    return NPERR_NO_ERROR;
}

int32_t pluginGetEdge(InstanceData* instanceData, RectEdge edge)
{
    // XXX nothing here yet since we don't support windowed plugins
    return NPTEST_INT32_ERROR;
}

int32_t pluginGetClipRegionRectCount(InstanceData* instanceData)
{
    // XXX nothing here yet since we don't support windowed plugins
    return NPTEST_INT32_ERROR;
}

int32_t pluginGetClipRegionRectEdge(InstanceData* instanceData,
                                    int32_t rectIndex, RectEdge edge)
{
    // XXX nothing here yet since we don't support windowed plugins
    return NPTEST_INT32_ERROR;
}

void pluginDoInternalConsistencyCheck(InstanceData* instanceData, string& error)
{
    //printf("pluginDoInternalConsistencyCheck\n");
}

static void
pluginDrawWindow(InstanceData* instanceData, void* event)
{
    NPWindow& window = instanceData->window;
    // When we have a widget, window.x/y are meaningless since our
    // widget is always positioned correctly and we just draw into it at 0,0
    int x = instanceData->hasWidget ? 0 : window.x;
    int y = instanceData->hasWidget ? 0 : window.y;
    int width = window.width;
    int height = window.height;

    QColor drawColor((QColor)instanceData->scriptableObject->drawColor);//QRgb qRgba ( int r, int g, int b, int a )
    QPixmap pixmap = QPixmap::fromX11Pixmap((XID)(((XGraphicsExposeEvent*)event)->drawable),QPixmap::ExplicitlyShared);

    if (instanceData->scriptableObject->drawMode == DM_SOLID_COLOR)
    {
        // drawing a solipixmapd color for reftests
        QPainter painter(&pixmap);
        painter.fillRect(QRect(x, y, width, height), drawColor);
        notifyDidPaint(instanceData);
        return;
    }

    NPP npp = instanceData->npp;
    if (!npp)
        return;

    QString text (NPN_UserAgent(npp));
    if (text.isEmpty())
        return;

    // drawing a in default grey with useragent string!
    QRect theRect(x, y, width, height);
    QColor color("#808080");
    //color.setAlpha(0);
    QBrush brush(color, Qt::SolidPattern);
    //printf("Opaque %d\n", brush.isOpaque());
    QPainter painter(&pixmap);
    QFont f("Helvetica");
    f.setPointSize(16);
    painter.setFont(f);

    painter.setCompositionMode(QPainter::CompositionMode_DestinationAtop);//CompositionMode_Clear CompositionMode_DestinationOver CompositionMode_SourceOver
    painter.fillRect(theRect, brush);
    painter.drawRect(theRect);
    painter.drawText(QRect(theRect), Qt::TextWordWrap, text);
    notifyDidPaint(instanceData);
    return;
}

static void
pluginDrawLocalWindow(InstanceData* instanceData, void* event)
{
#ifdef IMAGE_EXPOSE
    NPWindow& window = instanceData->window;
    // When we have a widget, window.x/y are meaningless since our
    // widget is always positioned correctly and we just draw into it at 0,0
    int x = instanceData->hasWidget ? 0 : window.x;
    int y = instanceData->hasWidget ? 0 : window.y;
    int width = window.width;
    int height = window.height;

    NPP npp = instanceData->npp;
    if (!npp)
        return;

    XGraphicsExposeEvent exposeEvent = ((XEvent*)event)->xgraphicsexpose;
    Display *disp = exposeEvent.display;
    if (disp) return;
    NPImageExpose *imgExp = (NPImageExpose*)exposeEvent.drawable;

/*
    printf(">>>>>>Func:%s::%d data:%p[%i,%i][bpl:%i][d:%i], r[%i,%i,%i,%i], tr[%g,%g], sc[%g,%g]\n", __FUNCTION__, __LINE__,
          imgExp->data, imgExp->dataSize.width, imgExp->dataSize.height, imgExp->stride, imgExp->depth,
          imgExp->x, imgExp->y, imgExp->width, imgExp->height,
          imgExp->translateX, imgExp->translateY, imgExp->scaleX, imgExp->scaleY);
*/

    QImage::Format format = QImage::Format_RGB32;
    if (imgExp->depth == 16)
        format = QImage::Format_RGB16;

    QImage image((unsigned char*)imgExp->data,
                 imgExp->dataSize.width,
                 imgExp->dataSize.height,
                 imgExp->stride,
                 format);

    QPainter p(&image);
    p.translate(imgExp->translateX, imgExp->translateY);
    p.scale(imgExp->scaleX, imgExp->scaleY);
    if (instanceData->isTransparent) {
        p.setCompositionMode(QPainter::CompositionMode_SourceOver);
        p.setOpacity(0.5);
    }
    QColor color((qrand() % 255), (qrand() % 255), (qrand() % 255));
    QBrush brush(color);

    p.fillRect(QRectF(imgExp->x, imgExp->y, imgExp->width, imgExp->height), brush);

    if (instanceData->showFps) {
        QRect theRect(x, y, width, height);
        QString text;
        text.setNum(instanceData->lastFpsNumber);
        QFont f("Helvetica");
        f.setPointSize(16);
        p.setFont(f);
        p.drawText(QRect(theRect), Qt::TextWordWrap, text);
    }
#else
    printf("Func:%s::%d attempt to call WindowlessLocal API\n", __PRETTY_FUNCTION__, __LINE__);
#endif

    return;
}
